BIOPAC_CONNECTION_CODES = {
    1: "MPSUCCESS",
    2: "MPDRVERR",
    3: "MPDLLBUSY",
    4: "MPINVPARA",
    5: "MPNOTCON",
    6: "MPREADY",
    7: "MPWPRETRIG",
    8: "MPWTRIG",
    9: "MPBUSY",
    10: "MPNOACTCH",
    11: "MPCOMERR",
    12: "MPINVTYPE",
    13: "MPNOTINNET",
    14: "MPSMPLDLERR",
    15: "MPMEMALLOCERR",
    16: "MPSOCKERR",
    17: "MPUNDRFLOW",
    18: "MPPRESETERR",
    19: "MPPARSEERR"
}        